import java.net.*;
import java.io.*;

/**
 * A thread to manage communication with one client. A LockManagerProtocol
 * will be created to treat the inputs.
 *
 * @author  Jyrki Nummenmaa
 * @version 0.1
 * @see     LockManager
 * @see     LockManagerProtocol
 */

public class LMMultiServerThread extends Thread
{
	private Socket socket = null;
	private LockManager lm = null;

	public LMMultiServerThread(LockManager lm, Socket socket)
	{
		super("LCMultiServerThread");
		this.socket = socket;
		this.lm = lm;
	}

	public void run()
	{

		try
		{
			PrintWriter out = new PrintWriter(socket.getOutputStream(), true);
			BufferedReader in =
				new BufferedReader(
					new InputStreamReader(socket.getInputStream()));

			String inputLine, outputLine;
			LockManagerProtocol lp = new LockManagerProtocol(lm);
			outputLine = lp.processInput(null);
			out.println(outputLine);

			while (true)
			{
				try
				{
					if (in.ready())
					{
						inputLine = in.readLine();
						
						/* TS 
						outputLine = lp.processInput(inputLine);
						out.println(outputLine); */
						ObjectOutputStream obOut;

						/* If queue information requested, open ObjectOutputStream
						 * and write locks ans lockQueue vectors from LockManager
						 * to the requestor.
						 */
						if(inputLine.equals("queue"))
						{
							obOut = new ObjectOutputStream(socket.getOutputStream());
							System.out.println(lm.serverId + " - Writing queue");
							obOut.writeObject(lm.getQueue());
							obOut.writeObject(lm.getLocks());
							outputLine = "Sent queue";
						}
						else
						{
							/* Otherwise do the original stuff */
							outputLine = lp.processInput(inputLine);
							out.println(outputLine);
						}
						/* TS */
						
						if (outputLine.equals("Bye"))
							break;
					}
					outputLine = lp.checkNotify();
					if (outputLine != null)
						out.println(outputLine);
					sleep(50);
				}
				catch (InterruptedException e)
				{
				}
			}
			out.close();
			in.close();
			socket.close();

		}
		catch (IOException e)
		{
			e.printStackTrace();
			System.out.println("Lost connection with a client");
		}
	}
}
