/*
 * java StartLockManager 33326 1 &
 * java StartLockManager 33327 2 &
 * java StartLockManager 33328 3 &
 * 
 * java StartLockClient 5 3 2 33335 localhost 33326 localhost 33327 localhost 33328 &
 * java LockClientController localhost 33335
 * 
 * java QueueClient localhost 33326 localhost 33327 localhost 33328
 */

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * A thread implementation of a LockClient. The LockClient will connect
 * the servers' server sockets and start listening on its own server socket.
 * When a ClientController connects its socket, it will start passing the 
 * ClientController's commands to the servers. It will output the response
 * from the servers. LockClient is started by a StartLockClient, which will
 * its argument array to LockClient - StartLockClients' usage is:
 * java StartLockClient txnId noXLockRequests noSLockRequests ownPort
 *    server1 port1 ... serverN portN
 * The Client will ask first noSLockRequests of parameter servers for an 
 * slock and last noXLockServers for an xlock.
 * 
 * @author  Jyrki Nummenmaa
 * @version 0.1
 * @see     StartLockClient
 * @see     LockManagerProtocol
 */

public class LockClient extends Thread
{

	int portNo;
	/** Server socket port for listening */
	int ownPortNo;
	/** txn id */
	/* TS
	int txnId = Integer.MAX_VALUE; */
	long txnId = -1;
	/** No of servers to consult for an xlock */
	int noXLockServers;
	/** No of servers to consult for an slock */
	int noSLockServers;
	/** No of server connections */
	int noConnections;
	/** Sockets for server connections */
	Socket[] lcSockets;
	/** PrintWriters for output to servers */
	PrintWriter[] lcOuts;
	/** BufferedReaders for inputs */
	BufferedReader[] lcIns;
	/** Argument vector */
	String[] args;
	/** For storing the active lock requests (granted or queued) */
	Vector lockRequests;
	

	public LockClient(String[] args)
	{
		this.args = args;
	}

	public void run()
	{
		try
		{

			noConnections = (args.length - 4) / 2;
			lcSockets = new Socket[noConnections];
			lcOuts = new PrintWriter[noConnections];
			lcIns = new BufferedReader[noConnections];

			/* TS
			txnId = Integer.parseInt(args[0]); */
			noXLockServers = Integer.parseInt(args[1]);
			noSLockServers = Integer.parseInt(args[2]);
			ownPortNo = Integer.parseInt(args[3]);

			lockRequests = new Vector(); // actice lock request type on item

			for (int i = 0; i < noConnections; i++)
			{
				portNo = Integer.parseInt(args[2 * i + 5]);
				lcSockets[i] = new Socket(args[2 * i + 4], portNo);
				lcOuts[i] =
					new PrintWriter(lcSockets[i].getOutputStream(), true);
				lcIns[i] =
					new BufferedReader(
						new InputStreamReader(lcSockets[i].getInputStream()));
			}

/* TS request TxnId from a randomly selected server */
			int tmp = (int)(Math.random() * noConnections);
			if(tmp == noConnections) tmp--;
			lcOuts[tmp].println("gettxnid");
			lcIns[tmp].readLine(); // For "Connection established"
			txnId = Long.parseLong(lcIns[tmp].readLine());			 
 /* TS */
 
			for (int i = 0; i < noConnections; i++)
			{
				lcOuts[i].println("txn " + txnId);
			}

		}
		catch (UnknownHostException e)
		{
			System.err.println("Don't know about host.");
			System.exit(1);
		}
		catch (IOException e)
		{
			System.err.println(
				"Couldn't get I/O for the connection to the host.");
			e.printStackTrace();
			System.exit(1);
		}
		catch (Exception e)
		{
			System.err.println(e);
			System.exit(1);
		}

		ServerSocket serverSocket = null;
		try
		{
			serverSocket = new ServerSocket(ownPortNo);
		}
		catch (IOException e)
		{
			System.err.println("Could not listen on port.");
			System.exit(1);
		}

		// Now we have set up the connections...wait for controller

		Socket clientSocket = null;
		BufferedReader controllerIn = null;
		try
		{
			clientSocket = serverSocket.accept();
			controllerIn =
				new BufferedReader(
					new InputStreamReader(clientSocket.getInputStream()));

		}
		catch (IOException e)
		{
			System.err.println("Accept failed.");
			System.exit(1);
		}

		try
		{

			String fromController = "";
			String fromServer = "";

			while (!fromController.equals("bye"))
			{
				for (int i = 0; i < noConnections; i++)
				{
					if (lcIns[i].ready())
					{
						fromServer = lcIns[i].readLine();
						System.out.println("Server " + i + ": " + fromServer);
					}
				}

				if (controllerIn.ready())
				{
					fromController = controllerIn.readLine();
					System.out.println("Controller: " + fromController);
					StringTokenizer st = new StringTokenizer(fromController);
					if (st.hasMoreTokens())
					{
						String oper = st.nextToken();
						if (st.hasMoreTokens())
						{
							int itemNo = Integer.parseInt(st.nextToken());
							if (oper.equals("slock"))
							{
								for (int i = 0; i < noSLockServers; i++)
								{
									lcOuts[i].println(fromController);
								}
								Lock l = new Lock(txnId, itemNo, Lock.SHARED);
								lockRequests.add(l);
							}
							if (oper.equals("xlock"))
							{
								for (int i = noConnections - 1;
									i >= (noConnections - noXLockServers);
									i--)
								{
									lcOuts[i].println(fromController);
								}
								Lock l =
									new Lock(txnId, itemNo, Lock.EXCLUSIVE);
								lockRequests.add(l);
							}
							if (oper.equals("unlock"))
							{
								boolean slockFound = false;
								boolean xlockFound = false;
								for (int i = 0; i < lockRequests.size(); i++)
								{
									Lock l = (Lock)lockRequests.elementAt(i);
									if ((l.getTxnId() == txnId)
										&& (l.getItemNo() == itemNo)
										&& (l.getLockType() == Lock.SHARED))
									{
										lockRequests.remove(i);
										if (!slockFound)
										{
											slockFound = true;
											for (int j = 0;
												j < noSLockServers;
												j++)
											{
												lcOuts[j].println(
													fromController);
											}
										}
									}
									if ((l.getTxnId() == txnId)
										&& (l.getItemNo() == itemNo)
										&& (l.getLockType() == Lock.EXCLUSIVE))
									{
										lockRequests.remove(i);
										if (!xlockFound)
										{
											xlockFound = true;
											for (int j = noConnections - 1;
												j
													>= (noConnections
														- noXLockServers);
												j--)
											{
												lcOuts[j].println(
													fromController);
											}
										}
									}
								}
							}
						}
						if(oper.equals("commit"))
						{
							int yes = 0, no = 0;
							for(int i = 0; i < noConnections; i++)
							{
								lcOuts[i].println("prepare");								
								if(lcIns[i].readLine().equals("yes"))
									yes++;
								else
									no++;
							}
							if(no == 0)
							{
								/* COMMIT */
								for(int i = 0; i < noConnections; i++)
								{
									lcOuts[i].println("commit");
								}								
							}
							else
							{
								/* ABORT */
								for(int i = 0; i < noConnections; i++)
								{
									lcOuts[i].println("abort");
								}								
							}
						}
					}
				}
			}

			for (int i = 0; i < noConnections; i++)
			{
				lcOuts[i].close();
				lcSockets[i].close();
			}

			controllerIn.close();
			clientSocket.close();
			serverSocket.close();

		}
		catch (Exception e)
		{
			System.out.println(e);
			e.printStackTrace();
		}

	}
}
