/*
 * Created on 9.11.2003
 *
 * File: QueueClient.java
 * 
 * The QueueClient asks for the lock and waits-for information from the
 * LockManager-servers and constructs a net from the information.
 * The net is not a global waits-for-graph, it visualices the lock-information.
 * The global-wfg can with quite small modifications be implemented and 
 * a cycle-seaching algorithm can then be performed in the graph.
 * 
 * To run the QueueClient type:
 * java QueueClient address1 port 1 ... serverN portN
 */

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * @author tpstrand
 *
 * Project: DTM
 */
public class QueueClient
{

	public static void main(String[] args)
	{
		Vector queues = new Vector(); 
		Vector locks = new Vector();
		int noConnections, portNo;
		Socket[] lcSockets;
		PrintWriter[] lcOuts;
		BufferedReader[] lcIns;
		ObjectInputStream objIn;

		noConnections = args.length / 2;
		lcSockets = new Socket[noConnections];
		lcOuts = new PrintWriter[noConnections];
		lcIns = new BufferedReader[noConnections];

		try
		{
			
			for (int i = 0; i < noConnections; i++)
			{
				portNo = Integer.parseInt(args[2 * i +1]);
				lcSockets[i] = new Socket(args[2 * i], portNo);
				lcOuts[i] =
					new PrintWriter(lcSockets[i].getOutputStream(), true);
				lcIns[i] = new BufferedReader(
					new InputStreamReader(lcSockets[i].getInputStream()));
				Thread.sleep(500);
				while(lcIns[i].ready())
				{
					lcIns[i].readLine();
					Thread.sleep(50);
				}
			}

			System.out.println("Sending queue..");
			Vector tmpVec;
			for (int i = 0; i < noConnections; i++)
			{
				lcOuts[i].println("queue");
				Thread.sleep(500); // To give the server time to open the ObjectStream
				objIn = new ObjectInputStream(lcSockets[i].getInputStream());
				tmpVec = (Vector)objIn.readObject();
				queues.add(tmpVec);
				tmpVec = (Vector)objIn.readObject();
				locks.add(tmpVec);
				System.out.println("Read queue Vector from server " + i);
				objIn.close();
			}	
			Vector vec = new Vector();
			Lock lc;							
			Net net = new Net();
			Node node;

			/* Locks */
			for(int i = 0; i < locks.size(); i++)
			{
				vec = (Vector)locks.elementAt(i);
				for(int j = 0; j < vec.size(); j++)
				{
					lc = (Lock)vec.elementAt(j);
					if(!net.nodes.containsKey(("" + lc.getTxnId())))
					{
						System.out.println("Adding " + lc.getTxnId() + " to net and lock " + lc.getItemNo());
						net.addNode(new Node(lc.getTxnId()));
						node = (Node)net.nodes.get(("" + lc.getTxnId()));
						node.addLock(lc.getItemNo());
					}
					else
					{
						System.out.println("Net already contains " + lc.getTxnId() + ", adding lock " + lc.getItemNo());
						node = (Node)net.nodes.get(("" + lc.getTxnId()));
						node.addLock(lc.getItemNo());
					}
				}
			}

			/* Queues */
			for(int i = 0; i < queues.size(); i++)
			{
				vec = (Vector)queues.elementAt(i);
				for(int j = 0; j < vec.size(); j++)
				{
					lc = (Lock)vec.elementAt(j);
					if(!net.nodes.containsKey(("" + lc.getTxnId())))
					{
						System.out.println("Adding " + lc.getTxnId() + " to net");
						net.addNode(new Node(lc.getTxnId()));
					}
					node = (Node)net.nodes.get(("" + lc.getTxnId()));
					if(!node.queues.containsKey(("" + lc.getTxnId())))
					{
						System.out.println("Adding " + lc.getItemNo() + " to queue");
						node.addQueue(lc.getItemNo());
					}
				}
			}

			System.out.println(net.toString());

			for (int i = 0; i < noConnections; i++)
			{
				lcOuts[i].close();
				lcIns[i].close();
				lcSockets[i].close();
			}

		}
		catch (UnknownHostException e)
		{
			System.err.println("Don't know about host.");
			System.exit(1);
		}
		catch (IOException e)
		{
			System.err.println(
				"Couldn't get I/O for the connection to the host.");
			e.printStackTrace();
			System.exit(1);
		}
		catch (Exception e)
		{
			System.err.println(e);
			e.printStackTrace();		
			System.exit(1);
		}
		
		
			
	}
}

class Net
{
	public Hashtable nodes = new Hashtable();
	
	public void addNode(Node node)
	{
		nodes.put(("" + node.id), node);
	}

	
	public String toString()
	{
		String ret = "\n\tNET\n\n";
		Enumeration enum = nodes.keys();
		Node node;
		
		while(enum.hasMoreElements())
		{
			String key = (String)enum.nextElement();
			ret += key + ":\n";
			node = (Node)nodes.get(key);
			ret += node.toString() + "\n\n";
		}
		return ret;
	}
}

class Node
{
	public long id;
	
	public Hashtable locks = new Hashtable();
	public Hashtable queues = new Hashtable();
	public boolean marked = false;
	
	public Node(long txnId)
	{
		id = txnId;
	}
	
	public void addLock(int i)
	{
		locks.put("" + i, new Integer(i));
	}

	public void addQueue(int i)
	{
		queues.put("" + i, new Integer(i));
	}
	
	public String toString()
	{
		String ret = "";
		Enumeration enum;
		
		enum = locks.keys();
		ret += "\tlocks: ";
		while(enum.hasMoreElements())
		{
			String key = (String)enum.nextElement();
			ret += key + ", ";
		}
		
		enum = queues.keys();
		ret += "\n\tqueues: ";
		while(enum.hasMoreElements())
		{
			String key = (String)enum.nextElement();
			ret += key + ", ";
		}
		return ret;
	}
}
		
		
		
		